@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995, 1996, 1997, 1998 Ben Laurie.  All rights reserved.
+ * Copyright (c) 1995, 1996, 1997, 1998, 1999 Ben Laurie.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -63,17 +63,24 @@
 #include <x509.h>
 #include <pem.h>
 #include <crypto.h>
+#include <rand.h>
 
 #include "httpd.h"
 #include "http_config.h"
 #include "http_log.h"
 #include "http_main.h"
 
+
+#if SSLEAY_VERSION_NUMBER < 0x0920
+# error "Don't use OpenSSL versions less than 0.9.2b, they have a serious security problem!"
+#endif
+
 /* Note that APACHE_SSL_EXPORT_CERTS is defined in buff.h */
 
 /* Temp */
 #define RENEG		FALSE
 #define DEBUG_RENEG	FALSE
+#define DEBUG_SSL_STATE FALSE
 
 #define USE_OLD_ENVIRONMENT	FALSE	/* Set to TRUE to use the old SSL_{CLIENT,SERVER}_I* instead of SSL_{CLIENT,SERVER}_I_* */
 
@@ -102,7 +109,7 @@
 #define FREE_SESSION	TRUE	/* Don't set this FALSE! */
 
 /* define this to debug the verification stuff */
-#define DEBUG_VERIFY	TRUE
+#define DEBUG_VERIFY	FALSE
 
 #if CACHE_SESSIONS
 # include "gcache.h"
@@ -110,7 +117,7 @@ static const char *s_szCacheServerPort;
 static const char *s_szCacheServerDir;
 static char *s_szCacheServerPath;
 
-#define DEBUG_SESSIONS		0
+# define DEBUG_SESSIONS	FALSE
 
 #endif
 
@@ -142,6 +149,13 @@ static char *s_szCacheServerPath;
 
 #define UNSET	-1
 
+/* Offsets for data stored in SSL structures */
+#define EX_DATA_CONN_REC	0
+#define EX_DATA_SERVER_REC	1
+#if RENEG
+# define EX_DATA_DIR_CONFIG	2
+#endif
+
 typedef enum
     {
     VERIFY_UNSET=-1,
@@ -171,8 +185,11 @@ typedef struct
     SSL_CTX *pSSLCtx;
 
     BOOL bFakeBasicAuth;
+    BOOL bNoCAList;
 
     time_t tSessionCacheTimeout;
+
+    array_header *aRandomFiles;
     } SSLConfigRec;
 
 typedef struct
@@ -195,11 +212,7 @@ typedef struct
 
 MODULE_VAR_EXPORT module apache_ssl_module;
 
-static conn_rec *s_pCurrentConnection;
 static pool *s_pPermanentPool;
-#if RENEG
-static SSLDirConfigRec *s_pConfig;
-#endif
 
 static const char six2pr[64+1]=
   "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
@@ -269,6 +282,70 @@ static void SSLLogErrorToFile(FILE *f)
 	}
     }
 
+#if DEBUG_SSL_STATE
+
+static void InfoCallback(SSL *s,int where,int ret)
+    {
+    char *str;
+    int w;
+    server_rec *pServer=SSL_get_ex_data(s,EX_DATA_SERVER_REC);
+
+    w=where&~SSL_ST_MASK;
+
+    if(w & SSL_ST_CONNECT)
+	str="SSL_connect";
+    else if(w & SSL_ST_ACCEPT)
+	str="SSL_accept";
+    else
+	str="undefined";
+
+    if(where & SSL_CB_LOOP)
+	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"%s:%s",
+		     str,SSL_state_string_long(s));
+    else if(where & SSL_CB_ALERT)
+	{
+	str=(where & SSL_CB_READ)?"read":"write";
+	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+		     "SSL3 alert %s:%s:%s",
+		     str,
+		     SSL_alert_type_string_long(ret),
+		     SSL_alert_desc_string_long(ret));
+	}
+    else if(where & SSL_CB_EXIT)
+	{
+	if(ret == 0)
+	    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+			 "%s:failed in %s",
+			 str,SSL_state_string_long(s));
+	else if (ret < 0)
+	    {
+	    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+			 "%s:error in %s",
+			 str,SSL_state_string_long(s));
+	    }
+	}
+    }
+
+#endif /* DEBUG_SSL_STATE */
+
+typedef struct
+    {
+    const char *szFile;
+    int nBytes;
+    } RandomFile;
+
+static void ReadRandomFiles(server_rec *pServer,RandomFile *pRF,int nRF)
+    {
+    int n;
+
+    for(n=0 ; n < nRF ; ++n,++pRF)
+	{
+	int r=RAND_load_file(pRF->szFile,pRF->nBytes);
+	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+		     "Random input %s(%d) -> %d",pRF->szFile,pRF->nBytes,r);
+	}
+    }
+
 #if APACHE_SSL_EXPORT_CERTS
 /* szEnv must have been pool allocated, or be static */
 static void AddCertToEnv(pool *p,table *e,char *szEnv,X509 *xs)
@@ -291,17 +368,23 @@ static int ApacheSSLVerifyCallback(int ok,X509_STORE_CTX *ctx)
     X509 *xs=X509_STORE_CTX_get_current_cert(ctx);
     int depth=X509_STORE_CTX_get_error_depth(ctx);
     int error=X509_STORE_CTX_get_error(ctx);
+    SSL *pSSL=X509_STORE_CTX_get_ex_data(ctx,
+					 SSL_get_ex_data_X509_STORE_CTX_idx());
+    conn_rec *pCurrentConnection=SSL_get_ex_data(pSSL,EX_DATA_CONN_REC);
 #else
 static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
 				   int error)
     {
 #endif
+
 #if RENEG
-    SSLDirConfigRec *pConfig=s_pConfig;
+    SSLDirConfigRec *pConfig=SSL_get_ex_data(pSSL,EX_DATA_DIR_CONFIG);
 #else
-    SSLConfigRec *pConfig=ap_get_module_config(s_pCurrentConnection->server->module_config,&apache_ssl_module);
+    SSLConfigRec *pConfig=
+      ap_get_module_config(pCurrentConnection->server->module_config,
+			   &apache_ssl_module);
 #endif
-    server_rec *pServer=s_pCurrentConnection->server;
+    server_rec *pServer=pCurrentConnection->server;
 
 #if DEBUG_VERIFY
       {
@@ -329,13 +412,13 @@ static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
     if(depth)
 	{
 	char *t;
-	pool *p=s_pCurrentConnection->client->pool;
+	pool *p=pCurrentConnection->client->pool;
 
-	if(!s_pCurrentConnection->client->tableCertChain)
-	    s_pCurrentConnection->client->tableCertChain=ap_make_table(p,5);
+	if(!pCurrentConnection->client->tableCertChain)
+	    pCurrentConnection->client->tableCertChain=ap_make_table(p,5);
 
 	t=ap_psprintf(p,ENV_CERT_CHAIN "%d",depth);
-	AddCertToEnv(p,s_pCurrentConnection->client->tableCertChain,t,xs);
+	AddCertToEnv(p,pCurrentConnection->client->tableCertChain,t,xs);
 	}
 #endif
 
@@ -371,8 +454,8 @@ static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
 	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,pServer,
 		      "verify error:num=%d:%s",error,
 		      X509_verify_cert_error_string(error));
-	SSLLogError(s_pCurrentConnection->server,APLOG_ERR,NULL);
-	s_pCurrentConnection->client->szClientX509=NULL;
+	SSLLogError(pCurrentConnection->server,APLOG_ERR,NULL);
+	pCurrentConnection->client->szClientX509=NULL;
 	}
     if(depth >= pConfig->nVerifyDepth)
 	{
@@ -407,8 +490,6 @@ static int VerifyFlags(SSLConfigRec *pConfig)
     default: break; /* avoid warnings from gcc */
 	}
 
-fprintf(stderr,"flags=%x\n",nVerify);
-
     return nVerify;
     }
 
@@ -419,9 +500,6 @@ static int SSLCheckCipher(request_rec *r)
       ap_get_module_config(r->per_dir_config,&apache_ssl_module);
     SSLConfigRec *pConfig=(SSLConfigRec *)
       ap_get_module_config(r->server->module_config,&apache_ssl_module);
-#if RENEG
-    int n;
-#endif
 
     if(rec->bSSLRequired && !r->connection->client->ssl)
 	{
@@ -439,13 +517,14 @@ static int SSLCheckCipher(request_rec *r)
 # endif
     if(1)
 	{
-	s_pConfig=rec;
+	SSL_set_ex_data(r->connection->client->ssl,EX_DATA_DIR_CONFIG,rec);
 	SSL_set_verify(r->connection->client->ssl,VerifyFlags(rec),
 		       ApacheSSLVerifyCallback);
 	SSL_renegotiate(r->connection->client->ssl);
 	SSL_do_handshake(r->connection->client->ssl);
-	SSL_write(r->connection->client->ssl,NULL,0);
-	SSL_read(r->connection->client->ssl,NULL,0);
+
+	r->connection->client->ssl->state=SSL_ST_ACCEPT;
+	SSL_do_handshake(r->connection->client->ssl);
 	}
 
     if(rec->nVerifyClient != VERIFY_NONE && rec->nVerifyClient != VERIFY_UNSET)
@@ -654,10 +733,9 @@ static int SSLFixups(request_rec *r)
 #if APACHE_SSL_EXPORT_CERTS
 # if !APACHE_SSL_KEEP_CERTS
     pCertStack=SSL_get_peer_cert_chain(r->connection->client->ssl);
-    /* XXX: When this works, check whether we need to start at 1, or add an offset */
-    for(n=1 ; pCertStack && n < sk_num(pCertStack) ; ++n)
+    for(n=0 ; pCertStack && n < sk_num(pCertStack) ; ++n)
 	{
-	t=ap_psprintf(r->pool,ENV_CERT_CHAIN "%d",n);
+	t=ap_psprintf(r->pool,ENV_CERT_CHAIN "%d",n+1);
 	AddCertToEnv(r->connection->client->pool,e,t,(X509 *)sk_value(pCertStack,n));
 	}
 # else
@@ -689,7 +767,9 @@ static void *CreateSSLServerConfig(pool *p,server_rec *s)
     rec->prsaKey=NULL;
     rec->pSSLCtx=NULL;
     rec->bFakeBasicAuth=UNSET;		/* Will become FALSE */
+    rec->bNoCAList=UNSET;
     rec->tSessionCacheTimeout=0;
+    rec->aRandomFiles=ap_make_array(p,1,sizeof(RandomFile));
 
     return rec;
     }
@@ -715,7 +795,9 @@ static void *MergeSSLServerConfig(pool *p,void *basev,void *addv)
     Merge(nVerifyClient,VERIFY_UNSET);
 #endif
     MergeBool(bFakeBasicAuth);
+    MergeBool(bNoCAList);
     MergeInt(tSessionCacheTimeout);
+    new->aRandomFiles=ap_append_arrays(p,base->aRandomFiles,add->aRandomFiles);
 
     return new;
     }
@@ -788,9 +870,9 @@ static void InitSSL(server_rec *s)
 #endif
     }
 
-#if SSLEAY_VERSION_NUMBER >= 0x0800
-/* FIXME: This is an expensive operation which should probably be done before
-forking */
+#if SSLEAY_VERSION_NUMBER >= 0x0800 && SSLEAY_VERSION_NUMBER < 0x0920
+/* FIXME: This ought to generate a new key from time to time - why make the
+Feds' life easy? */
 static RSA *TmpRSACallback(SSL *pSSL,int nExport)
     {
     static RSA *pRSA=NULL;
@@ -805,6 +887,28 @@ static RSA *TmpRSACallback(SSL *pSSL,int nExport)
     }
 #endif
 
+#if SSLEAY_VERSION_NUMBER >= 0x0920
+/* FIXME: This ought to generate a new key from time to time - why make the
+Feds' life easy? */
+static RSA *TmpRSACallback(SSL *pSSL,int nExport,int nKeyLength)
+    {
+    static RSA *pRSA512=NULL;
+    static RSA *pRSA1024=NULL;
+
+    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,
+		 SSL_get_ex_data(pSSL,EX_DATA_SERVER_REC),
+		 "Generating %d bit key",nKeyLength);
+    assert(nExport);  /* Otherwise it was a sign-only key and we should set our own keylength */
+    assert(nKeyLength == 512 || nKeyLength == 1024);
+    if (pRSA512 == NULL && nKeyLength == 512)
+	pRSA512=RSA_generate_key(512,RSA_F4,NULL,NULL);
+    if (pRSA1024 == NULL && nKeyLength == 1024)
+	pRSA1024=RSA_generate_key(1024,RSA_F4,NULL,NULL);
+    
+    return nKeyLength == 512 ? pRSA512 : pRSA1024;
+    }
+#endif
+
 #if CACHE_SESSIONS
 
 #define MAX_SESSION_DER	100000
@@ -847,7 +951,7 @@ static uchar *CacheFind(uchar *aucKey,int nKey,int *pnData,BOOL *pbLocal)
     return t;
     }
 
-#if APACHE_SSL_EXPORT_CERTS
+#if APACHE_SSL_KEEP_CERTS
 static uchar *MakeCertKey(Cache *pCache,const char *szKey,int *pnLength)
     {
     uchar *buf=malloc(*pnLength=pCache->nKey+strlen(szKey));
@@ -876,16 +980,16 @@ static int SendCertToServer(void *extra,const char *szKey,const char *szValue)
 static uchar *CacheFindS(char *aucKey,int nKey,int *pnData,BOOL *pbLocal)
     { return CacheFind((uchar *)aucKey,nKey,pnData,pbLocal); }
 
-static void SendCertChainToServer(Cache *pCache)
+static void SendCertChainToServer(Cache *pCache,conn_rec *pCurrentConnection)
     {
-    if(!s_pCurrentConnection->client->tableCertChain)
+    if(!pCurrentConnection->client->tableCertChain)
 	return;
     ap_table_do(SendCertToServer,pCache,
-		s_pCurrentConnection->client->tableCertChain,NULL);
+		pCurrentConnection->client->tableCertChain,NULL);
     }
 
 static void GetCertChainFromServer(uchar *aucSessionID,int nLength,
-				   BOOL bLocal)
+				   BOOL bLocal,conn_rec *pCurrentConnection)
     {
     int n;
     char buf[MAX_SESSION_DER];
@@ -905,17 +1009,18 @@ static void GetCertChainFromServer(uchar *aucSessionID,int nLength,
 	if(!t)
 	    break;
 
-	if(!s_pCurrentConnection->client->tableCertChain)
-	    s_pCurrentConnection->client->tableCertChain=
-	      ap_make_table(s_pCurrentConnection->client->pool,5);
+	if(!pCurrentConnection->client->tableCertChain)
+	    pCurrentConnection->client->tableCertChain=
+	      ap_make_table(pCurrentConnection->client->pool,5);
 
-	ap_table_set(s_pCurrentConnection->client->tableCertChain,
+	ap_table_set(pCurrentConnection->client->tableCertChain,
 		     buf+nLength,(char *)t);
 	}
     }
-#endif
+#endif /* APACHE_SSL_KEEP_CERTS */
 
-static void SendSessionToServer(SSL_SESSION *pSession)
+static void SendSessionToServer(SSL_SESSION *pSession,
+				conn_rec *pCurrentConnection)
     {
     uchar buf[MAX_SESSION_DER];
     Cache *p;
@@ -927,7 +1032,7 @@ static void SendSessionToServer(SSL_SESSION *pSession)
     if(LocalCacheFind(pSession->session_id,pSession->session_id_length))
 	return;
 
-    pConfig=ap_get_module_config(s_pCurrentConnection->server->module_config,
+    pConfig=ap_get_module_config(pCurrentConnection->server->module_config,
 				 &apache_ssl_module);
     tExpiresAt=time(NULL)+pConfig->tSessionCacheTimeout;
 
@@ -945,20 +1050,21 @@ static void SendSessionToServer(SSL_SESSION *pSession)
     p->nData=nLength;
 
     GlobalCacheAdd(p->aucKey,p->nKey,p->aucData,p->nData,tExpiresAt);
-#if APACHE_SSL_EXPORT_CERTS
-    SendCertChainToServer(p);
+#if APACHE_SSL_KEEP_CERTS
+    SendCertChainToServer(p,pCurrentConnection);
 #endif
     }
 
-static SSL_SESSION *GetSessionFromServer(uchar *aucSessionID,int nLength)
+static SSL_SESSION *GetSessionFromServer(uchar *aucSessionID,int nLength,
+					 conn_rec *pCurrentConnection)
     {
     SSL_SESSION *pSession;
     uchar *t;
     int nData;
     BOOL bLocal=FALSE;
 
-#if APACHE_SSL_EXPORT_CERTS
-    GetCertChainFromServer(aucSessionID,nLength,bLocal);
+#if APACHE_SSL_KEEP_CERTS
+    GetCertChainFromServer(aucSessionID,nLength,bLocal,pCurrentConnection);
 #endif
 
     t=CacheFind(aucSessionID,nLength,&nData,&bLocal);
@@ -984,11 +1090,13 @@ static int NewSessionCacheCallback(SSL *pSSL,SSL_SESSION *pNew)
     fprintf(stderr,")\n");
 #endif
 
-    SendSessionToServer(pNew);
+    SendSessionToServer(pNew,SSL_get_ex_data(pSSL,EX_DATA_CONN_REC));
     return 0;
     }
 
-static SSL_SESSION *GetSessionCacheCallback(SSL *pSSL,unsigned char *aucSessionID,int nLength,int *pCopy)
+static SSL_SESSION *GetSessionCacheCallback(SSL *pSSL,
+					    unsigned char *aucSessionID,
+					    int nLength,int *pCopy)
     {
     SSL_SESSION *pSession;
 
@@ -1001,7 +1109,8 @@ static SSL_SESSION *GetSessionCacheCallback(SSL *pSSL,unsigned char *aucSessionI
     fprintf(stderr,")\n");
 #endif
     *pCopy=0;
-    pSession=GetSessionFromServer(aucSessionID,nLength);
+    pSession=GetSessionFromServer(aucSessionID,nLength,
+				  SSL_get_ex_data(pSSL,EX_DATA_CONN_REC));
 
     return pSession;
     }
@@ -1061,15 +1170,20 @@ static void GetPrivateKey(server_rec *s,SSLConfigRec *pConfig,
 		  (const char *)pConfig->prsaKey);
     }
 
+#if SSLEAY_VERSION_NUMBER < 0x0920 || RENEG
 static int X509NameCompare(X509_NAME **a,X509_NAME **b)
     {
-    return(X509_NAME_cmp(*a,*b));
+    return X509_NAME_cmp(*a,*b);
     }
+#endif
 
 static void GetCertificateAndKey(server_rec *s,pool *p,SSLConfigRec *pConfig)
     {
     FILE *f;
     char szPath[MAX_STRING_LEN];
+#if SSLEAY_VERSION_NUMBER >= 0x0920
+    STACK *stackCAs=NULL;
+#endif
 
     if(pConfig->px509Certificate)
 	{
@@ -1162,10 +1276,41 @@ static void GetCertificateAndKey(server_rec *s,pool *p,SSLConfigRec *pConfig)
 
 #if SSLEAY_VERSION_NUMBER >= 0x0800
     SSL_CTX_set_tmp_rsa_callback(pConfig->pSSLCtx,TmpRSACallback);
+#endif
 
+#if SSLEAY_VERSION_NUMBER >= 0x0920
+#if RENEG
+    if(!pConfig->bNoCAList)
+#else
+    if(pConfig->nVerifyClient != VERIFY_OPTIONAL_NO_CA && !pConfig->bNoCAList)
+#endif
+	{
+	if(pConfig->szCACertificateFile || pConfig->szCACertificatePath)
+	    stackCAs=sk_new(NULL);
+
+	if(pConfig->szCACertificateFile
+	   && !SSL_add_file_cert_subjects_to_stack(stackCAs,
+					      pConfig->szCACertificateFile))
+	    {
+	    SSLLogError(s,APLOG_CRIT,"error reading CA certs");
+	    exit(1);
+	    }
+
+	if(pConfig->szCACertificatePath
+	   && !SSL_add_dir_cert_subjects_to_stack(stackCAs,
+					     pConfig->szCACertificatePath))
+	    {
+	    SSLLogError(s,APLOG_CRIT,"error reading CA certs");
+	    exit(1);
+	    }
+
+	if(stackCAs && sk_num(stackCAs) > 0)
+	    SSL_CTX_set_client_CA_list(pConfig->pSSLCtx,stackCAs);
+	}
+#elif SSLEAY_VERSION_NUMBER >= 0x0800
     if(pConfig->szCACertificateFile)
 	SSL_CTX_set_client_CA_list(pConfig->pSSLCtx,
-				   SSL_load_client_CA_file(pConfig->szCACertificateFile));
+		SSL_load_client_CA_file(pConfig->szCACertificateFile));
 
     if(pConfig->szCACertificatePath)
 	{
@@ -1239,7 +1384,8 @@ static void LaunchCache(server_rec *s,pool *pPool)
 
     s_pServer=s;
     printf("Launching... %s\n",s_szCacheServerPath);
-    pid=ap_spawn_child(pPool,CacheLauncher,s_szCacheServerPath,kill_always,NULL,NULL,NULL);
+    pid=ap_spawn_child(pPool,CacheLauncher,s_szCacheServerPath,kill_always,
+		       NULL,NULL,NULL);
     printf("pid=%d\n",pid);
     }
 #endif
@@ -1286,6 +1432,8 @@ static void InitSSLServer(server_rec *s,pool *p)
 #endif
 	if(pConfig->bFakeBasicAuth == UNSET)
 	    pConfig->bFakeBasicAuth=FALSE;
+	if(pConfig->bNoCAList == UNSET)
+	    pConfig->bNoCAList=FALSE;
 
 	/* End of fixups */
 
@@ -1317,7 +1465,8 @@ static void InitSSLServer(server_rec *s,pool *p)
 	    exit(1);
 	    }
 #if !RENEG
-	if(pConfig->nVerifyClient < 0 || pConfig->nVerifyClient > VERIFY_OPTIONAL_NO_CA)
+	if(pConfig->nVerifyClient < 0
+	   || pConfig->nVerifyClient > VERIFY_OPTIONAL_NO_CA)
 	    {
 	    ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,s,
 			 "Bad value for SSLVerifyClient (%d)",
@@ -1414,8 +1563,6 @@ static void ApacheSSLSetupVerify(conn_rec *conn)
 
     conn->client->szClientX509=NULL;
 
-    s_pCurrentConnection=conn;
-
     conn->client->nVerifyError=X509_V_OK;
 
 #if SSLEAY_VERSION_NUMBER < 0x0800
@@ -1442,18 +1589,17 @@ static void CloseSSL(void *_conn)
     SSL_free(ssl);
     /*    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,conn->server,
 		 "Closed");*/
-#if RENEG
-    s_pCurrentConnection=NULL;
-#endif
     }
 #endif
 
-int ApacheSSLSetupConnection(conn_rec * conn)
+int ApacheSSLSetupConnection(conn_rec *conn)
     {
     server_rec *srvr=conn->server;
     BUFF *fb=conn->client;
     SSLConfigRec *pConfig=ap_get_module_config(srvr->module_config,
 					       &apache_ssl_module);
+    int r;
+
     if(pConfig->bDisabled)
 	{
 	fb->ssl=NULL;
@@ -1461,17 +1607,32 @@ int ApacheSSLSetupConnection(conn_rec * conn)
 	}
 
     ap_conn_timeout("SSL accept timeout",conn);
-    
-    s_pCurrentConnection=conn;
+
+    ReadRandomFiles(srvr,(RandomFile *)pConfig->aRandomFiles->elts,
+		    pConfig->aRandomFiles->nelts);
 
 #if APACHE_SSL_KEEP_CERTS
     fb->tableCertChain=NULL;
 #endif
+#if SSLEAY_VERSION_NUMBER >= 0x0920
+    /* Use our current configuration as the session ID context. Do this by
+       actually using its address, which should be invariant */
     fb->ssl=SSL_new(pConfig->pSSLCtx);
+    r=SSL_set_session_id_context(fb->ssl,(unsigned char *)&pConfig,
+				 sizeof pConfig);
+    ap_assert(r);
+#else
+    fb->ssl=SSL_new(pConfig->pSSLCtx);
+#endif
     SSL_set_fd(fb->ssl,fb->fd);
 #if FREE_SESSION
     ap_register_cleanup(conn->pool,conn,CloseSSL,CloseSSL);
 #endif
+#if DEBUG_SSL_STATE
+    SSL_set_info_callback(fb->ssl,InfoCallback);
+    SSL_set_ex_data(fb->ssl,EX_DATA_SERVER_REC,srvr);
+#endif
+    SSL_set_ex_data(fb->ssl,EX_DATA_CONN_REC,conn);
 
     ApacheSSLSetupVerify(conn);
 
@@ -1516,7 +1677,8 @@ int ApacheSSLSetupConnection(conn_rec * conn)
 		}
 	    }
 
-	if(pConfig->nVerifyClient == VERIFY_REQUIRE && !conn->client->szClientX509)
+	if(pConfig->nVerifyClient == VERIFY_REQUIRE
+	   && !conn->client->szClientX509)
 	    {
 	    SSLLogError(conn->server,APLOG_ERR,"No client certificate");
 	    return 0;
@@ -1529,8 +1691,6 @@ int ApacheSSLSetupConnection(conn_rec * conn)
 #if !RENEG
     /* This should be safe.... so I'll use it */
     SSL_set_read_ahead(fb->ssl,1);
-
-    s_pCurrentConnection=NULL;
 #endif
 
     return TRUE;
@@ -1612,6 +1772,31 @@ static const char *SSLCacheServerDir(cmd_parms *cmd,char *struct_ptr,
     }
 #endif
 
+static const char *SSLRandomFile(cmd_parms *cmd,char *struct_ptr,char *arg1,
+				 char *arg2)
+    {
+    RandomFile rf;
+
+    rf.szFile=arg1;
+    rf.nBytes=atoi(arg2);
+    ReadRandomFiles(cmd->server,&rf,1);
+
+    return NULL;
+    }
+
+static const char *SSLRandomFilePerConnection(cmd_parms *cmd,char *struct_ptr,
+					      char *arg1,char *arg2)
+    {
+    SSLConfigRec *pConfig=ap_get_module_config(cmd->server->module_config,
+						  &apache_ssl_module);
+    RandomFile *prf=ap_push_array(pConfig->aRandomFiles);
+
+    prf->szFile=arg1;
+    prf->nBytes=atoi(arg2);
+
+    return NULL;
+    }
+
 static const char *set_server_string_slot (cmd_parms *cmd,char *struct_ptr,
 					   char *arg)
     {
@@ -1683,6 +1868,9 @@ command_rec ssl_cmds[]=
     { "SSLCACertificateFile",set_server_string_slot,
       (void *)XtOffsetOf(SSLConfigRec,szCACertificateFile),RSRC_CONF,TAKE1,
       "CA Certificate file (taken from SSL_CERT_FILE if absent)" },
+    { "SSLNoCAList",set_server_bool_slot,
+      (void *)XtOffsetOf(SSLConfigRec,bNoCAList),RSRC_CONF,NO_ARGS,
+      "Don't send the CA list to the client" },
 #if RENEG
     { "SSLVerifyDepth",set_dir_int_slot,
       (void *)XtOffsetOf(SSLDirConfigRec,nVerifyDepth),OR_FILEINFO,TAKE1,
@@ -1725,6 +1913,11 @@ command_rec ssl_cmds[]=
     { "SSLCacheServerRunDir",SSLCacheServerDir,NULL,RSRC_CONF,TAKE1,
       "Path for the server to run in (in case of coredumps)" },
 #endif
+    { "SSLRandomFile",SSLRandomFile,NULL,RSRC_CONF,TAKE2,
+      "A file to read randomness from and a maximum number of bytes to read" },
+    { "SSLRandomFilePerConnection", SSLRandomFilePerConnection,NULL,RSRC_CONF,
+      TAKE2,
+      "A file to read randomness from and a maximim number of bytes to read on each request" },
     { "SSLLogFile",SSLLogFile,NULL,RSRC_CONF,TAKE1,
       "Defunct SSL log file - DO NOT USE" },
     { NULL },
